\ 3iForth.fs -- 3-Instruction Forth for AVR Atmel 328 / Arduino Nano on GForth 0.7.0 (Windows PC)
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ 2020-06-22

\ See Article & Discussion at http://www.mathscitech.org/articles/3iForth
\ TARGET = 3iForth.avr running on ATmel328 in Arduino Nano (compiled & uploaded to Nano using free Arduino IDE & default bootloader)
\ HOST = 3iForth.fs running in GForth 0.7.0 on Windows PC
\ Nano on startup (or reset) sets up serial port (9600 baud), initializes data stack, and waits for instructions over serial line

\ c. 250 lines of Forth code

\ USAGE:  
\   1. Connect Arduino Nano to HOST using USB mini cable
\           PINS - note: A0 through A7 are pins 14 to 21.  A0 to A5 (14-19) are I/O, A6/A7 (20-21) are AnalogRead only.
\   2. Run RealTerm with command line setup: baud=9600 half port=6.  This sets up the serial ports correctly to use 
\   3. HOST > run Gforth > include 3iForth.fs
\   4. HOST > help, help1, help2   (provides instructions)
\ Demo board 1: LED board with 11 LEDs alternating red and yellow (pins 2-12) and a tiny speaker (pin 13)
\ Demo board 2: Programmable Digital Synthesizer board with 13 keys and speaker

( * SERIAL PORT - these words open and close a given COM port on the PC * )
(   ex:            5 open-com              )
(   MISSING:  need to add GForth code for configuring the COM port.  Workaround: Use RealTerm to connect first  )

decimal
\ interactivity
$1B constant ESC   
: pause key drop ; \ press any key  \ loop till key pressed can be done with begin ... key? until 

\ logic
: not ( 0/x -- 1/0 ) if 0 else 1 then ;
: not4 ( $xx -- $yy ) invert $F and ; \ 4-bit complement
: not8 ( $xx -- $yy ) invert $FF and ; \ 4-bit complement

variable COMhandle
: save-handle ( handle -- ) COMhandle ! ;
: fh ( -- handle ) COMhandle @ ;

: com-string ( port# -- c-adr cnt )
    \ Builds COM port string.  Test: 6 com-string type
    >r s" comX" 2dup + 1 - '0' r> + swap c! ; 
: com-open? ( handle errcode -- )
    dup 0= ." ..." 
    if drop save-handle ." COM port opened ... " 
    else cr ." **error: check device is plugged into correct COM port" cr throw
    then ;
:  [OPEN-COM] ( port# --  )  \ Opens COM port in binary mode
    com-string r/w bin create-file com-open? 1500 ms ." OK" ;  \ Settling time delay of 1000ms doesn't work, 1500ms works
:  CLOSE-COM ( -- )
    fh close-file throw cr ." COM port closed." cr ;

( * MEMORY  - these words are essential for working efficiently with memory buffers, i.e. arrays of memory * )
1 constant byte 
2 constant wword
4 constant dword
: bytes  byte * ;
: wwords wword * ;   
: dwords dword * ;   \ GForth is 32-bits
\ Efficient fetch & store from/into and an array
\ BYTES
:  c@+ ( adr -- adr' val )  
    dup 1 bytes + swap    c@ ;
:  c!+ ( val adr -- adr' )  
    dup 1 bytes + -rot    c! ;

: go-to-end  ( array narray sz -- array-end narray ) 
    swap >r r@ * + 1 - r> ;
:  c@- ( adr -- adr' val )  
    dup 1 bytes - swap    c@ ;
\ WORDS (2 bytes)
\ ADD w, to complement c,
: bytes2word ( hi8 lo8 -- val16 )
	swap    8 lshift   or ;
: lohi ( val16 -- hi8 lo8  )  \ convert val16 to little endian byte stream ( low on top of stack) ex: $FFFE lohi  ( -- $FF $FE )
    dup   $FF00 and   8 rshift    swap $FF and ;
    \ 440 f2hl 1 184
: ww@ ( adr -- val16 )  \ little endian retrieval
	c@+    swap c@    swap  bytes2word ;
: ww@+  ( adr -- adr' val16 )  
	c@+    swap c@+   swap -rot  swap  bytes2word ;
: ww!  ( val16 adr -- )  \ little endian storage
	swap lohi rot  c!+ c! ;
: ww!+ ( val16 adr -- adr' )  
	swap lohi rot  c!+ c!+ ;
\ DWORDS (4 bytes)
: @+ 
    dup 1 dwords + swap @ ;
: !+
    dup 1 dwords + -rot ! ;

\ STACK HELP
: 3drop 2drop drop ;
: 4drop 2drop 2drop ;  \ for 4-bit port work
: ndrop ( n-elements N -- ) 0 do drop loop ;  \ needed in 3iShift.fs beats sequencer
: ndup  ( n-elements N --  n-elements  n-elements ) dup 0 do dup pick swap loop drop ; \ needed in 3iShift.fs beats sequencer
: nrev  ( a1 a2 a3 .. aN  N -- aN aN-1 .. a3 a2 a1 )  \ needed in various places (search *reverse*) 
   1 do i roll loop ;          \ 3iPlaylist_book.fs music playback, 3iFont.fs, 3iI2C.fs, 3iArt.fs
   


\ math
: /up   ( a b -- ceil[a/b]  )   /mod swap 0 > if 1 + then ;  \ integer divide & round up     
: /dn   ( a b -- floor[a/b] )   / ;


\ characters
: is_between? ( n a b -- t/f ) rot dup rot > -rot > or invert ;  \ not ( n>b or n<a )
: ascii  255 0 do i 
    dup 14 = if 11 emit space 12 emit space then
    dup 4 mod 0 = if 2 spaces dup 8 mod 0 = if cr then then 
    emit space loop 
    cr ." 7 = bell" 
    cr ." 8 = ?"
    cr ." 9 = space"
    cr ." 10 = newline"
    cr ." 13 = newline"
;

( * BINARY viewing bitfield of a number  - determines number of bits from T )
:  T   $FF  ;  \ 1111b  Determines whether calculations are 4-, 8-, 16-bits, or 32-bits
\ :  T   $FFFF ;
\ :  T   $FFFFFFFF ; \ 32 bits  T = -1
:  ~    ( p -- ~p ) T xor ;          \ negation op takes bitfield size from T
: .bit  ( n -- ) 0> if '1' else '0' then emit ;
: init-mask  T 2 / ~  ;  \ initial mask is $8=1000b, $80=1000000b, or $8000 depending on T
: bits  (  bitfield  --  )  \  loop through bits from MSB mask and print
    init-mask begin 2dup and .bit 1 rshift dup 0= until 2drop ." b" ;


: .b ( bf -- bf ) dup bits 2 spaces ;
: .sh ( -- ) hex .s decimal ;  \ show stack in hex

: >b8 ( 8-bits -- byte ) 0 8 1 do or swap i lshift loop or ;  \ test: 219 .b  1 1 0 1  1 0 1 1  >b8 .
\ : b8>



( * SERIAL READ - gets messages from target into byte receive buffer.  From there read using c@ * )

create RCV-BUFFER 255 bytes allot  \ buffer for holding messages from the uC
: show-rcv ( cnt -- cnt )   \  ex: 255 show-rcv
    \ Prints rcv buffer in decimal  ( use hex. instead of . to show values in hex  )
    dup RCV-BUFFER swap dup 0> 
    if 0 cr ." <rcv: " do c@+ . ." " loop else drop then drop ;    
: read-com ( n -- cnt )   \ read N bytes  ex: 50 read-com show-rcv
    RCV-BUFFER swap fh read-file throw ( -- cnt ) ;
\ : bytes. ( cnt -- cnt ) dup ." >> " . ." bytes on port" ; 

\ COMPLICATION: for REAL-TIME read, i.e. after sending a serial command and expecting a response from the uC on Serial.  80 ms works, 60 intermittent, 25 doesn't work
: rxN ( N -- cnt ) 80 ms read-com ( bytes. ) show-rcv ;     
: rx ( -- ) 50 rxN drop ;

( * SERIAL TRANSMIT - send messages in prepared message buffer to target * )
( replaced pad with msg-buffer so we retain control of the scratch space, no risk of volatile pad and can inspect anytime ) 
( * SENDING REMOTE INSTRUCTIONS TO TARGET - some work with parameters that must reside on the target's data-stack * )
\ Notes: 
\ 1.   tx ;      only xcall eventually transits.  the others prepare the MSG-BUFFER (fewer comms -> more efficient)
\ 2.   -- adr'   each function accepts an address and provides the next address for easy chaining into MSG-BUFFER
create MSG-BUFFER 20 bytes allot     \  write buffer
: show-msg ( cnt -- )                \ ex: 10 show-msg
    MSG-BUFFER swap 0 cr ." MSG-BUFFER>: " do c@+ . ." " loop drop ;          \ change . to hex. to show hex values
: [tx] ( msg-buf len -- )  \ serial transmit len bytes from msg-buf
    fh write-file throw fh flush-file throw  ; 
: tx  ( adr' -- )  MSG-BUFFER swap over -  [tx] ;  \ used with [memstore] and [nstore] words
: tx1 ( byte -- )  MSG-BUFFER c!   MSG-BUFFER 1  [tx] ;  \ send 1 byte on serial


\ These are 3iForth commands, as distinct from exported chip instructions
\ Have a remote stack allows testing chip remotely using local stack as if it were a local development, and without risking pins etc
0 constant CMD_X_S  \ maps to stack_len() in 3iForth.avr
1 constant CMD_XSTORE
2 constant CMD_XDROP
3 constant CMD_XFETCH
4 constant CMD_XCALL

\ 0-parameter commands
: xfetch  ( -- val )  CMD_XFETCH tx1 80 ms 1 read-com drop rcv-buffer c@ ;  \ fetch top byte from remote stack & return to TOS local
: x_s     ( -- )      cr CMD_X_S tx1 rx xfetch ." (". ." )" ;  \ queries chip and reports size of stack in bytes
: xdrop   ( -- )      CMD_XDROP tx1 ;           \ remove a byte from remote stack
: 2xdrop  xdrop xdrop ;
: 4xdrop  2xdrop 2xdrop ;

\ for paramaterized commands, we use msg-buffer and the machinery below

\ Note: in all cases except for unrestricted byte transfers ( eg 3iI2C.avr cmdOLED and dataOLED words) we keep the same stack diagram in host and target (.fs and .avr).  
\ Because serial transmit does so from the start of a buffer (FIFO) we need to reverse the host stack to insert in right order
\ array storage also in FIFO order matching stack and then from there into target machine (2 LIFO = FIFO).
\ stores into host memory  in LIFO order as  cmd  wn wn-1 ... w1 - ONLY HERE
\ Use MSG-BUFFER before [memstore]/[nstore] or >r/r> if appending to an existing store operation
: [memstore]  ( a adr -- adr' )  \ 1 param storage.  xstore does the storage directly (without memstore)  
    CMD_XSTORE swap c!+ c!+ ;  
: [nstore]  ( w1 w2 ... wn adr n -- adr' ) \ 
    swap over 0 do over i - 1 + roll swap [memstore] loop swap drop ;
: [revstore] ( w1 w2 ... wn adr n -- adr' ) \ load into memory in reverse order, wn first
    0 do [memstore] loop ;
: nstore   ( n -- )  msg-buffer swap [nstore]   tx ;  \ transmit to target
: revstore ( n -- )  msg-buffer swap [revstore] tx ;  \ transmit to target

\ 1-parameter commands
: xstore  ( a -- )    1 nstore  ;      \ store byte onto remote stack using msg-buffer
\ Note: the parameter is trailing in the serial queue so that when xstore is recognized, it pulls the next value from the queue.
: xcall   ( instr -- ) xstore CMD_XCALL tx1 ;     \ call export chip instruction (see .avr fcall jump table)

: open-com ( usb# -- ) [open-com] 5 rxN drop ;

( * HOST INTERFACE  CALL - on the host machine we don't need to know the mechanism, we think we live in the target and call the exported functions directly using Forth * )

\ WARNING: these match the jumptable entries in the chip
0 constant  CALL_PINMODE
1 constant  CALL_PINWRITE
2 constant  CALL_DELAY
3 constant  CALL_TONE
4 constant  CALL_SETREG
5 constant  CALL_SWITCHKEY  \ digital piano only (piano.avr and piano-4x4.avr)
6 constant  CALL_RECORD     \ digital piano only (piano.avr and piano-4x4.avr)
7 constant  CALL_PINREAD
8 constant  CALL_KEYPAD     \ keypad only (see 3iSensor.fs)
9 constant  CALL_PINREADLEVEL
$A constant CALL_PINPWM
$B constant CALL_PORT4MODE
$C constant CALL_PORT4WRITE
$D constant CALL_PORT4READ
$E constant CALL_CHECKREG
\ $F through $16 are  8 instructions related to use of OLED displays (oled.avr)
\ $17 is motor movement instruction (motor.avr)
$18 constant CALL_SETTIMER  \ changes PWM timer
$19 constant CALL_PINSIGNAL
$1A constant CALL_PORT4SIGNAL
\ $1B pollpianokey - piano.fs
$1C constant CALL_SHIFTOUTBYTE
$1D constant CALL_SHIFTOUT32
$1E constant CALL_SHIFTINBYTE
$1F constant CALL_I2C_INIT
$20 constant CALL_I2C_SETCLOCK
$21 constant CALL_I2C_CMDOLED
$22 constant CALL_I2C_DATAOLED


1 constant OUTPUT
0 constant INPUT
2 constant INPUT_PULLUP
\ NO built-in Arduina constant for INPUT_PULLDOWN

: pinMode ( mode pin -- )  2 nstore CALL_PINMODE xcall 60 ms ;   
    \ ex: 1 9 pinmode  (output)  modes: 0x0 INPUT, 0x1 OUTPUT, 0x2 INPUT_PULLUP (when active low buttons)
    \ Note: need settle time on pin when changing mode.  60 ms is magic number
: out ( pin -- )   OUTPUT swap pinmode ;
: in ( pin -- )    INPUT swap pinmode ;
: pullup ( pin -- ) INPUT_PULLUP swap pinmode ;  \ ex: 2 pullup
: pullup-keys  \ set all keyboard pins 2-12 to INPUT_PULLUP
    13 2 do i pullup loop 
    \ NEED TO ALSO DO A0 and A1 -- not sure how to pass those parameters!
    100 ms ; 

1 constant HIGH_SIGNAL
0 constant LOW_SIGNAL

: pinwrite ( level pin -- )  2 nstore  CALL_PINWRITE xcall ;  \ ex: 1 9 pinwrite   0 9 pinwrite
: high ( pin -- )  HIGH_SIGNAL swap pinwrite ;
: low ( pin -- )   LOW_SIGNAL swap pinwrite ;
: lo low ;
: hi high ;
: testpin ( n -- )  dup dup out high 500 ms low ." done." ;

: pwm ( val8 pin -- ) 2 nstore CALL_PINPWM xcall ;   \ 8-bit value (0-255)  pins 3,9,10,11 (490 Hz), pins 5,6 (980 Hz)
: testpwm ( -- ) 50 10 pwm 500 ms  255 10 pwm 500 ms 0 10 pwm 500 ms ;
\ max is -212, max before hitting stop is -208, min is -2, 0 is weird

: settimer ( prescale timer -- ) 2 nstore CALL_SETTIMER xcall ;  \ Timer 0 pins 5,6,  Timer 1 pins 9,10, Timer 2 pins 3, 11

\ FIX NEEDED: ms is usually more than 1 byte worth
: delay ( byte -- ) xstore CALL_DELAY xcall ; \ ex: 128 delay   NEEDED on TARGET to preserve timings when running untethered

: [tone] ( freq-hi freq-lo dur-hi dur-lo pin -- ) 5 nstore CALL_TONE xcall ;
   \ 13 setspeaker
   \ 440 lohi 500 lohi speaker [tone] 600 ms 440 lohi 500 lohi speaker [tone]  \ ringing tone! :)
   \ https://www.arduino.cc/reference/en/language/functions/advanced-io/tone/
   \ https://github.com/bhagman/Tone#ugly-details

: setreg ( val16 reg -- )   \ ex: 440 0 setreg    sets register0 to 16-bit value   ( e.g. musical note A4=440 Hz )
    swap lohi rot   3 nstore CALL_SETREG xcall ;
: checkreg ( reg -- serial: hi lo ) xstore CALL_CHECKREG xcall rx ;

	
: pinread ( pin -- 0/1 )   \ digitalRead() - with realtime delay (COMPLICATION)  -- WARNING: set pin mode to INPUT (active high) or PULLUP (active low)
	xstore CALL_PINREAD xcall
	120 ms   \ 50 ms seems to work ( 0 failures out of 30 trials ), if processor is also writing to display, then 120ms is enough
	1 read-com if rcv-buffer c@ else cr ." FAIL pinread: no bytes" cr then ;
: poll-pin ( pin -- ) begin dup pinread . key? until ." Done!" drop key drop ;   \ for active low.  if active high remove the not
: pinsignal ( pin -- ) \ read until get a signal
    xstore CALL_PINSIGNAL xcall
    begin 100 ms 1 read-com until rcv-buffer c@ ;  \ BLOCKING READ

: pinreadLevel ( Apin -- counts )   \ analogRead() -- NOTE: use ANALOG pin (A0-A7, 14-21) - result in 10 bit counts (0 to 1024)
    \ WARNING: set pin mode to INPUT
    \ Ref: https://www.arduino.cc/reference/en/language/functions/analog-io/analogread/
	xstore CALL_PINREADLEVEL xcall
	60 ms   \ 50 ms seems to work ( 0 failures out of 30 trials )
	2 read-com if rcv-buffer c@+ 8 lshift swap c@ + else cr ." FAIL pinread: no bytes" cr then ;
: Vref5 ( cnts -- mV ) 49 10 */ ;  \ 4.9mV per count  (5V=1024 counts)
: poll-level ( Apin -- ) begin dup pinreadLevel Vref5 . cr 100 ms key? until ." Done!" drop ;
: plot-level ( Apin -- ) begin dup pinreadLevel Vref5 dup 100 / cr spaces ." XX  " . 100 ms key? until ." Done!" drop ;


( * SETTING UP AND WORKING WITH 4-BIT PORTS --- IN PROGRESS 2020-09-22---Tue---07:52 * )
15 constant REG_PORTAOUT
16 constant REG_PORTBIN
17 constant REG_PORTCOUT
18 constant REG_PORTDIN
\ Physical Port Control   Ex:  5 4 3 2  pins for output port, lsb TOS.   Ex: 11 10 9 8 pins for input port, lsb TOS
: join4 ( msb c b lsb -- mcbl-16bits )  \  ex: 5 4 3 2 -- $5432 (in hex) combines 4 bytes (pins) to 16-bit value for setreg
    swap 4 lshift or
    swap 8 lshift or
    swap $C lshift or ;

\ CODE REVIEW: consolidate and adjust in all .fs user code
: port4out ( d c b a -- )  \ ex: 5 4 3 2  port4out, lsb TOS      Test: reg_portout checkreg   WARNING: MAX PIN IS 15
    join4 REG_PORTAOUT setreg 100 ms
    OUTPUT REG_PORTAOUT 2 nstore CALL_PORT4MODE xcall ;
: port4Cout ( d c b a -- ) \ lsb at TOS   WARNING: MAX PIN IS 15
    join4 REG_PORTCOUT setreg 100 ms
    OUTPUT REG_PORTCOUT 2 nstore CALL_PORT4MODE xcall ;
: port4write  ( hex -- ) REG_PORTAOUT 2 nstore CALL_PORT4WRITE xcall ;           \ writes hex digit to 4-bit port
: port4Cwrite ( hex -- ) REG_PORTCOUT 2 nstore CALL_PORT4WRITE xcall ;           \ writes hex digit to 4-bit port


: port4in  ( d c b a -- )  \ ex: 11 10 9 8  port4in, lsb TOS     Test: reg_portin checkreg
    join4  REG_PORTBIN  setreg 80 ms
    INPUT REG_PORTBIN 2 nstore CALL_PORT4MODE xcall ;
: port4read ( -- hex )         \ reads hex digit from 4-bit port
    REG_PORTBIN  2 nstore CALL_PORT4READ  xcall 
    100 ms  \ 60 ms seems to work
    1 read-com if rcv-buffer c@ else cr ." FAIL port4read: no bytes" cr then ;
: poll-port ( -- ) begin port4read not4 dup if . cr else ." ." drop then key? until ." Done!" ;  \ use not4 for active low
: port4signal ( -- ) \ polls input port until something is received (on serial)
    REG_PORTBIN xstore CALL_PORT4SIGNAL xcall
    begin 100 ms 1 read-com until rcv-buffer c@ ; \ BLOCKING READ
    
\ SPECIAL IC Chips
\ 8-bits shift out register IC 74595  -- shift.fs

: pulse-hi ( pin -- ) dup hi 1 ms lo ;   \ rising edge pulse (for clk and latch)
: pulse-lo ( pin -- ) dup lo 1 ms hi ;   \ falling edge pulse (for clear)



( * PROBLEM ORIENTED LANGUAGE - ON THE HOST --- setting up a simplified yet powerful custom dictionary * )

: reload ( -- )   \ ensure com port is closed before reloading which clobbers fh so need to open-com port again
  fh 0> if close-com then
  s" 3iForth.fs" included ;

( * HELP FILES * )

: help
    cr
    cr ." help "
    cr ."  ------------------------------------------------------------"
    cr ."  3-instruction Forth for Atmel 328 chip:"
    cr ."     STARTING/STOPPING"
    cr ."       1 open-com   rx   -  opens communication to port 1 and flushes response (should be 255 254)"
    cr ."       close-com - to end session"
    cr ."       reload - to reload 3iForth.fs file after instruction edits"
    cr ."     OUTPUT PINS"
	cr ."       2 out         - sets pin 2 into output mode"
    cr ."       2 high        - sends 5V to pin 2  (choose pins 2 through 12)"
    cr ."       2 low         - sends 0V to pin 2  (choose pins 2 through 12)"
	cr ."	 INPUT PINS"
    cr ."       2 pullup     - sets pin 2 into pullup high mode, i.e. pin is then active low"
	cr ."       2 pinread "
	cr ."    LOWEST LEVEL => help1"
    cr ."       host:     x_s   1 xstore   xdrop  xfetch .   <instr> xcall   "
    cr ."       target:   ms  pinmode   pinwrite   pinread "
    cr ."                 pinwriteLevel pinreadLevel  setreg  checkreg  "
    cr ."                 port4mode  port4write  port4read  tone  "
    cr ."    Add-ons after 3iForth.fs:"
    cr ."       1/ include   3iSound-4x4.fs"
    cr ."       2/ include   3iShift.fs   3iPianoLights.fs   3iRecord.fs  3iPlaylist_book.fs   3iGinan.fs"
    cr ."       3/ include   3iOLED.fs    3iSensor.fs"
    cr ."       4/ include   3iBeats.fs   3iBeats_cookbook.fs   3iMenu.fs"
    cr ."       5/ include   3iMotor.fs   3iRadio.fs"
    cr ."       obsolete:    3iPort-4x4.fs, 3iLight.fs"
    cr cr
;

: help1 ( -- )
    cr ."  2.  1  13 pinmode       In general:  pinMode ( mode pin --  )"
    cr ."  2.  2   9 pinmode       set pin 9 to INPUT_PULLUP"
    cr ."  3.  1  13 pinwrite        In general:  pinwrite  ( level pin -- )   (called digitalWrite in Arduino C)."
    cr ."  4.  0  13 pinwrite        "
;

\ ======================
help


\    cr
\    cr ."   => LEDs & BUZZERS:  include 3iLight.fs"
\    cr ."       all-out  all-high all-low  5 n-vivid
\    cr ."   => SOUND / MUSIC (tone)   include 3iSound.fs"  
\    cr ."       start-speaker  - sets pin 13 to speaker" 


\ FOR VERSION HISTORY see 3iForth_history.txt file
